# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

# pylint: disable=missing-module-docstring
# pylint: disable=missing-function-docstring

import ast
from dataclasses import dataclass
import hashlib
import logging
from pathlib import Path
from pprint import pformat
from textwrap import dedent as tw_dedent, indent as tw_indent, wrap as tw_wrap
from typing import List, Optional
from .types import (
    ImportedModule,
    ImportedModules,
    ModuleSource,
    ModuleSourceSearchFunc,
    SourceText,
)

log = logging.getLogger(__name__)


@dataclass
class _PossibleImport:
    name: str
    relative_level: int
    line_num: int
    known_module: bool


def _find_possible_imports_in_module(source: ModuleSource) -> List[_PossibleImport]:
    """
    Parse a module's source code and extract a list of possible modules and package imported by a module.  In many cases, imported
    items are explicitly known to be a module or package. but in some cases inspection outside of the module source is required.
    For instance, given the statement "from foo import bar", bar might be a module but that is not possible to determine in the
    context of just the source code.
    """
    tree = ast.parse("\n".join(source.text), str(source.path))
    possible_imports: List[_PossibleImport] = []
    for node in ast.walk(tree):
        if isinstance(node, ast.Import):
            for alias in node.names:
                possible_imports.append(_PossibleImport(alias.name, 0, alias.lineno, True))
        elif isinstance(node, ast.ImportFrom):
            base_module_name = ""
            if node.module:
                possible_imports.append(_PossibleImport(node.module, node.level, node.lineno, True))
                base_module_name = node.module + "."
            for alias in node.names:
                possible_imports.append(_PossibleImport(base_module_name + alias.name, node.level, alias.lineno, False))
    return possible_imports


def get_module_source(path: Path, module_fullname: Optional[str] = None) -> ModuleSource:
    """
    Read in the source code for a Python module identified by a pathlib.Path object.  Strip all whitespace at the end of each line,
    and return it as an array of strings.
    """
    with path.open() as source_fo:
        text = [s.rstrip() for s in source_fo.readlines()]
        return ModuleSource(text, path, module_fullname)


def quote_source_text(source_text: SourceText) -> SourceText:
    quoted_source_text = ['r"""']
    for line in source_text:
        quoted_source_text.append(line.replace('"""', '""" + \'"""\' + r"""'))
    quoted_source_text.append(r'"""')
    return quoted_source_text


def indent_source_text(source_text: SourceText, indent=4) -> SourceText:
    indent_str = " " * indent
    return [(indent_str + line).rstrip() for line in source_text]


def hash_source_text(source_text: SourceText) -> str:
    source_hash = hashlib.sha256()
    for line in source_text:
        source_hash.update(line.encode())
    return source_hash.hexdigest()


def format_comments(text: str, width=79, indent=0) -> SourceText:
    indent_str = " " * indent + "# "
    return tw_wrap(
        tw_dedent(text).rstrip(),
        width=width,
        initial_indent=indent_str,
        subsequent_indent=indent_str,
    )


def get_imported_modules(source: ModuleSource) -> ImportedModules:
    """
    Given source text for a Python module, return a dictionary containing the imported modules indexed by the module's fullname.
    """
    imported_modules: ImportedModules = {}
    source_module_layers = source.module_fullname and source.module_fullname.split(".") or []
    for pi in _find_possible_imports_in_module(source):
        import_module_name = pi.name
        import_module_layers = import_module_name and import_module_name.split(".") or []
        if pi.relative_level:
            import_module_rel_cnt = pi.relative_level - (1 if source.path and source.path.name == "__init__.py" else 0)
            if not source.path or import_module_rel_cnt >= len(source_module_layers):
                raise ImportError(
                    f"attempted relative import with no known parent package on line {pi.line_num} of {str(source.path)}",
                    name=import_module_name,
                )
            import_module_layers[0:0] = source_module_layers[: (-import_module_rel_cnt or None)]

        # log.debug(
        #     f"{import_module_rel_name=} {import_module_rel_cnt=} {import_module_name=} {import_module_layers=} "
        #     f"{source_module_layers=}"
        # )

        # FIXME: TRACK IMPORTS FROM MULTIPLE SOURCE FILES AND LINES WITHIN A SOURCE FILE
        # add all layers of the imported module to set of imported modules
        import_module_fullname = ""
        for import_module_layer in import_module_layers:
            import_module_fullname += import_module_layer
            if import_module_fullname not in imported_modules:
                imported_modules[import_module_fullname] = ImportedModule(import_module_fullname, source, pi.line_num)
            import_module_fullname += "."

    if log.isEnabledFor(logging.DEBUG):
        # pylint: disable=logging-fstring-interpolation
        msg_txt = tw_indent("\n".join(source.text), "    ").rstrip()
        log.debug(f"Source code for {str(source.path)}:\n{msg_txt}")
        msg_txt = tw_indent(pformat(imported_modules, sort_dicts=False), "    ")
        log.debug(f"Modules imported by {str(source.path)}:\n{msg_txt}")

    return imported_modules


def get_all_imported_modules(source: ModuleSource, module_source_search_fn: ModuleSourceSearchFunc) -> ImportedModules:
    all_imported_modules: ImportedModules = {}
    unprocessed_source = [source]
    while unprocessed_source:
        next_source = unprocessed_source.pop()
        imported_modules = get_imported_modules(next_source)
        for imported_module in imported_modules.values():
            if imported_module.fullname not in all_imported_modules:
                imported_module.path = module_source_search_fn(imported_module.fullname)
                if imported_module.path:
                    imported_module.source = get_module_source(imported_module.path, imported_module.fullname)
                    unprocessed_source.append(imported_module.source)
                all_imported_modules[imported_module.fullname] = imported_module
    return all_imported_modules
