# Copyright (C) 2024, UChicago Argonne, LLC
# Licensed under the 3-clause BSD license.  See accompanying LICENSE.txt file
# in the top-level directory.

"""Supporting Serialization additions for json."""
import datetime
import decimal

import numpy as np
import pandas as pd
import simplejson as json
from dateutil.parser import parse as date_parse

from Octeres.forthwith import FORMAT_DATE_FULL_TZ, FORMAT_DATE_FULL, FORMAT_DATE_DAY

# cannot import, must directly obtain it.
_NoneType = type(None)


class DataEncoder(json.JSONEncoder):
    """An extension to the JSONEncoder that allows for serializing
    data that was not provided by default"""

    other_encoders = []

    def __init__(self, *args, **kwargs):
        """Init for the Encoder.
        This modifies two variables to allow custom handling of
        Decimal and tuples."""
        kwargs["tuple_as_array"] = False  # HACKZ
        kwargs["use_decimal"] = False  # HACKZ
        self.np_allowed_dtypes = [
            "int8",
            "int16",
            "int32",
            "int64",
            "uint8",
            "uint16",
            "uint32",
            "uint64",
        ]
        super(DataEncoder, self).__init__(*args, **kwargs)

    @classmethod
    def register(cls, ocls, func):
        cls.other_encoders.append((ocls, func))

    def default(self, obj):  # pylint: disable=E0202
        """The encoder"""
        if isinstance(obj, tuple):
            return {"_type_": "tuple", "_value_": list(obj)}
        if isinstance(obj, decimal.Decimal):
            return {"_type_": "Decimal", "_value_": str(obj)}
        if isinstance(obj, datetime.datetime):
            if obj.tzinfo is not None:
                return {
                    "_type_": "datetime.datetime tz",
                    "_value_": obj.strftime(FORMAT_DATE_FULL_TZ),
                }
            else:
                return {
                    "_type_": "datetime.datetime",
                    "_value_": obj.strftime(FORMAT_DATE_FULL),
                }
        if isinstance(obj, datetime.date):
            return {"_type_": "datetime.date", "_value_": obj.strftime(FORMAT_DATE_DAY)}
        if isinstance(obj, np.uint8):
            return {"_type_": "numpy.uint8", "_value_": str(obj)}
        if isinstance(obj, np.uint16):
            return {"_type_": "numpy.uint16", "_value_": str(obj)}
        if isinstance(obj, np.uint32):
            return {"_type_": "numpy.uint32", "_value_": str(obj)}
        if isinstance(obj, np.uint64):
            return {"_type_": "numpy.uint64", "_value_": str(obj)}
        if isinstance(obj, np.int8):
            return {"_type_": "numpy.int8", "_value_": str(obj)}
        if isinstance(obj, np.int16):
            return {"_type_": "numpy.int16", "_value_": str(obj)}
        if isinstance(obj, np.int32):
            return {"_type_": "numpy.int32", "_value_": str(obj)}
        if isinstance(obj, np.int64):
            return {"_type_": "numpy.int64", "_value_": str(obj)}
        if isinstance(obj, np.ndarray):
            dtype = obj.dtype.name
            if dtype in self.np_allowed_dtypes:
                return {
                    "_type_": "numpy.ndarray",
                    "_value_": obj.tolist(),
                    "_subtype_": dtype,
                }
        if isinstance(obj, type):
            # let's be specific about the types we support
            # it also should probably be only primitives
            if obj == int:
                return {"_type_": "type", "_value_": obj.__name__}
            if obj == str:
                return {"_type_": "type", "_value_": obj.__name__}
            if obj == list:
                return {"_type_": "type", "_value_": obj.__name__}
            if obj == tuple:
                return {"_type_": "type", "_value_": obj.__name__}
            if obj == _NoneType:
                return {"_type_": "type", "_value_": obj.__name__}
            if obj == dict:
                return {"_type_": "type", "_value_": obj.__name__}
            if obj == bool:
                return {"_type_": "type", "_value_": obj.__name__}
            if obj == datetime.datetime:
                return {"_type_": "type", "_value_": obj.__name__}
        for cls, func in self.other_encoders:
            if isinstance(obj, cls):
                return {"_type_": obj.__class__.__name__, "_value_": func(obj)}
        return super(DataEncoder, self).default(obj)


class DataDecoder(json.JSONDecoder):
    """This will decode a string encoded by DataEncoder."""

    other_decoders = []

    def __init__(self, *args, **kwargs):
        """Init for the Decoder."""
        self.np_allowed_dtypes = [
            "int8",
            "int16",
            "int32",
            "int64",
            "uint8",
            "uint16",
            "uint32",
            "uint64",
        ]
        self.__special_keys__ = {"_type_", "_value_"}
        super(DataDecoder, self).__init__(*args, **kwargs)

    @classmethod
    def register(cls, cls_str, func):
        cls.other_decoders.append((cls_str, func))

    def extra_type_parser(self, obj):
        """This is the decoder function helper.  This is a possible point of
        slowness as it has to walk through all the structures."""
        if isinstance(obj, list):
            new_lst = []
            for sub_obj in obj:
                new_lst.append(self.extra_type_parser(sub_obj))
            result = new_lst
        elif isinstance(obj, dict):
            result = None
            if set(obj.keys()).issuperset(self.__special_keys__):
                obj_type = obj["_type_"]
                obj_value = obj["_value_"]
                if obj_type == "Decimal":
                    result = decimal.Decimal(obj_value)
                elif obj_type == "tuple":
                    objs = tuple(obj_value)
                    new_lst = []
                    for sub_obj in objs:
                        new_lst.append(self.extra_type_parser(sub_obj))
                    result = tuple(new_lst)
                elif obj_type == "datetime.datetime":
                    result = datetime.datetime.strptime(obj_value, FORMAT_DATE_FULL)
                elif obj_type == "datetime.datetime tz":
                    # python is very weak here, the easiest option is date_parse
                    result = date_parse(obj_value)
                elif obj_type == "datetime.date":
                    result = datetime.datetime.strptime(obj_value, FORMAT_DATE_DAY).date()
                elif obj_type == "numpy.uint8":
                    result = np.uint8(obj_value)
                elif obj_type == "numpy.uint16":
                    result = np.uint16(obj_value)
                elif obj_type == "numpy.uint32":
                    result = np.uint32(obj_value)
                elif obj_type == "numpy.uint64":
                    result = np.uint64(obj_value)
                elif obj_type == "numpy.ndarray":
                    obj_subtype = obj["_subtype_"]
                    if obj_subtype in self.np_allowed_dtypes:
                        dtype = np.dtype(obj_subtype)
                        result = np.array(obj_value, dtype=dtype)
                elif obj_type == "type" and obj_value == "str":
                    result = str
                elif obj_type == "type" and obj_value == "int":
                    result = int
                elif obj_type == "type" and obj_value == "tuple":
                    result = tuple
                elif obj_type == "type" and obj_value == "list":
                    result = list
                elif obj_type == "type" and obj_value == "NoneType":
                    result = _NoneType
                elif obj_type == "type" and obj_value == "dict":
                    result = dict
                elif obj_type == "type" and obj_value == "bool":
                    result = bool
                elif obj_type == "type" and obj_value == "datetime":
                    result = bool
                else:
                    result = None
                    for obj_type_str, func in self.other_decoders:
                        if obj_type == obj_type_str:
                            result = func(obj_value)
                    if result is None:
                        result = obj
            else:
                new_dct = {}
                for key, sub_obj in obj.items():
                    new_dct[key] = self.extra_type_parser(sub_obj)
                result = new_dct
        else:
            result = obj
        return result

    def decode(self, json_str, **kwargs):
        """This is the decoder function."""
        obj = super().decode(json_str, **kwargs)
        obj = self.extra_type_parser(obj)
        return obj


def flatten_dct_types(dct):
    """given a dictionary, given specific types, convert them
    to a type that will be easier to use in another language such as javascript.
    These should be used for display only due to the lossiness method of flattening."""
    new_dct = {}
    for key, value in dct.items():
        type_value = type(value)
        if type_value in (datetime.datetime, pd.Timestamp):
            new_dct[key] = value.strftime(FORMAT_DATE_FULL)
        elif type_value == decimal.Decimal:
            new_dct[key] = str(value)
        elif type_value == tuple:
            new_dct[key] = ",".join(value)
        elif type_value in (np.int64, np.uint8):
            new_dct[key] = int(value)
        else:
            new_dct[key] = value
    return new_dct


def flatten_lstofdct_types(lstofdct):
    """given a list of dictionaries, given specific types, convert them
    to a type that will be easier to use in another language such as javascript.
    These should be used for display only due to the lossiness method of flattening."""
    new_lstofdct = []
    for dct in lstofdct:
        new_dct = flatten_dct_types(dct)
        new_lstofdct.append(new_dct)
    return new_lstofdct


def octeres_dumps(obj: object):
    return json.dumps(obj, cls=DataEncoder)


def octeres_loads(data: str):
    return json.loads(data, cls=DataDecoder)
